<?php

namespace Drupal\sso_trigger\EventSubscriber;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\PageCache\ResponsePolicy\KillSwitch;
use Drupal\Core\Path\CurrentPathStack;
use Drupal\Core\Path\PathMatcher;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Session\AnonymousUserSession;
use Drupal\Core\Session\SessionManagerInterface;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Url;
use Drupal\autologout\AutologoutManagerInterface;
use Drupal\simplesamlphp_auth\Service\SimplesamlphpAuthManager;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Cookie;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Event subscriber subscribing to KernelEvents::REQUEST.
 */
class RedirectAnonymousSubscriber implements EventSubscriberInterface {
  use LoggerChannelTrait;
  use MessengerTrait;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $account;

  /**
   * The CurrentPathStack service.
   *
   * @var \Drupal\Core\Path\CurrentPathStack
   */
  protected $pathStackService;

  /**
   * The Path Matcher service.
   *
   * @var \Drupal\Core\Path\PathMatcher
   */
  protected $pathMatcherService;

  /**
   * The SimpleSAML PHP Auth Manager.
   *
   * @var \Drupal\simplesamlphp_auth\Service\SimplesamlphpAuthManagerq
   */
  protected $authManager;

  /**
   * The Autologout manager.
   *
   * @var \Drupal\autologout\AutologoutManagerInterface
   */
  protected $authlogoutManager;

  /**
   * The session manager.
   *
   * @var \Drupal\CoreSessionSessionManagerInterface
   */
  protected $sessionManager;

  /**
   * The page cache kill switch.
   *
   * @var \Drupal\Core\PageCache\ResponsePolicy\KillSwitch
   */
  protected $pageCacheKillSwitch;

  /**
   * The request path.
   *
   * Stored as a property to avoid continually accessing Drupal's
   * path.current service.
   *
   * @var string
   */
  protected $requestPath;

  /**
   * The path that triggers single sign-on (SSO).
   *
   * @var string
   */
  protected $ssoPath;

  /**
   * Paths for which we are configured to not trigger SSO.
   *
   * @var string
   */
  protected $excludedPaths;

  /**
   * Config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Create a RedirectAnonymousSubscriber object.
   *
   * @param \Drupal\Core\Session\AccountProxyInterface $account
   *   The account to use.
   * @param \Drupal\Core\Path\CurrentPathStack $path_current
   *   Path Stack Service.
   * @param \Drupal\Core\Path\PathMatcher $path_matcher
   *   Path Matcher service.
   * @param \Drupal\simplesamlphhp_auth\Service\SimplesamlphpAuthManager $auth_manager
   *   The SAML Authentication Manager.
   * @param \Drupal\Core\PageCache\ResponsePolicy\KillSwitch $page_cache_kill_switch
   *   The cache kill switch service.
   * @param \Drupal\autologout\AutologoutManagerInterface $autologout_manager
   *   The Autologout manager.
   * @param \Drupal\Core\Session\SessionManagerInterface $session_manager
   *   The session manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\TempStore\PrivateTempStoreFactory $tempstore_factory
   *   A tempstore factory.
   */
  public function __construct(AccountProxyInterface $account, CurrentPathStack $path_current, PathMatcher $path_matcher, SimplesamlphpAuthManager $auth_manager, KillSwitch $page_cache_kill_switch, AutologoutManagerInterface $autologout_manager, SessionManagerInterface $session_manager, ConfigFactoryInterface $config_factory, PrivateTempStoreFactory $tempstore_factory) {
    $this->account = $account;
    $this->pathStackService = $path_current;
    $this->pathMatcherService = $path_matcher;
    $this->authManager = $auth_manager;
    $this->pageCacheKillSwitch = $page_cache_kill_switch;
    $this->autologoutManager = $autologout_manager;
    $this->sessionManager = $session_manager;
    $this->configFactory = $config_factory;
    $this->tempStoreFactory = $tempstore_factory;
  }

  /**
   * Callback to check the authenticated status of the request.
   *
   * If not authenticated, and if the requested URL is not one of the excepted
   * paths, then redirect to the SSO trigger path.
   */
  public function checkAuthStatus(RequestEvent $event) {
    $sso_config = $this->configFactory->get('sso_trigger.settings');
    if (!$sso_config) {
      $this->getLogger('sso_trigger')->notice('Could not get config');
      return;
    }

    // If we're not configured to redirect anonymous visitors to the sso path,
    // then do nothing.
    if (!$sso_config->get('enabled')) {
      return;
    }

    // Save some path values for future use.
    // @todo Move these if-statements to helper getters.
    if (!isset($this->requestPath)) {
      $this->requestPath = $this->pathStackService->getPath();
    }

    if (!isset($this->ssoPath)) {
      $this->ssoPath = $sso_config->get('sso_path');
    }

    $this->excludedPaths = $sso_config->get('excluded_paths');
    if (!is_array($this->excludedPaths)) {
      $this->excludedPaths = [];
    }

    // Path-based checks.
    if ($this->shouldDoNothing()) {
      // $this->getLogger('sso_trigger')
      // ->notice('Returning without doing anything');
      return;
    }

    // If we're anonymous, then
    // convert the requested path into a ReturnTo query parameter and
    // redirect to the SSO path.
    if ($this->account->isAnonymous()) {
      // Check to see if we've already authenticated, but the user is blocked.
      if ($this->authManager->isAuthenticated()) {
        $authname = $this->authManager->getAuthname();
        $account = user_load_by_name($authname);
        if ($account && !empty($account->id()) && !$account->isActive()) {
          // The user is blocked, so redirect to /user/login with an informative
          // message.
          $this->getLogger('sso_trigger')->notice('User @name is blocked', ['@name' => $authname]);

          $blocked_user_message = $sso_config->get('blocked_user_message');
          $message = check_markup($blocked_user_message['value'], $blocked_user_message['format']);
          $this->messenger()->addError($message);

          // Construct our redirect URL.
          $goto = Url::fromUri('internal:/user/login', ['absolute' => FALSE])->toString();
          $this->getLogger('sso_trigger')->notice('Redirecting to @goto', ['@goto' => $goto]);
          $response = new RedirectResponse($goto, RedirectResponse::HTTP_FOUND);

          // Prevent caching this page response (may be unnecessary)
          $this->pageCacheKillSwitch->trigger();
          $event->setResponse($response);
          // $this->authManager->logout('/user/login');
          return;
        }
      }

      $this->triggerSso($event);
      return;
    }
    else {
      $this->mitigateAutoLogout($event);
    }
  }

  /**
   * Should this module should avoid taking action for the current request path.
   */
  protected function shouldDoNothing() {
    // First, check to see if this is a request for system.cron
    // Can't use RouteMatch - maybe because we're running too early in the
    // bootstrap.
    $path_array = explode('/', $this->requestPath);
    if (!empty($path_array[1]) && $path_array[1] == 'cron') {
      return TRUE;
    }

    // Next check if this request has anything to do with the autologout module.
    if (!empty($path_array[1]) && strpos($path_array[1], 'autologout') === 0) {
      return TRUE;
    }

    // Check against the main sso trigger path & our configured excluded paths.
    if ($this->requestPath == $this->ssoPath || $this->pathMatcherService->matchPath($this->requestPath, implode("\n", $this->excludedPaths))) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * Construct and set a redirect response to trigger an SSO login.
   */
  protected function triggerSso(RequestEvent $event) {
    // Set a default return_to value.
    $return_to = NULL;
    $front_path = $this->configFactory->get('system.site')->get('page.front');

    if ($this->requestPath != '/' && $this->requestPath != $front_path) {
      $return_to = $this->requestPath;
    }

    // Construct our redirect URL.
    $goto = Url::fromUri('internal:' . $this->ssoPath, ['absolute' => FALSE])->toString();

    $this->getLogger('sso_trigger')->notice('Redirecting to @goto', ['@goto' => $goto]);

    $response = new RedirectResponse($goto, RedirectResponse::HTTP_FOUND);

    if (!empty($return_to)) {
      $this->getLogger('sso_trigger')->notice('Setting ssoReturnTo cookie to @t', ['@t' => $return_to]);
      $cookie = new Cookie('ssoReturnTo', $return_to, 0, "", "", TRUE, TRUE);
      $response->headers->setCookie($cookie);
    }

    // Prevent caching this page response (may be unnecessary)
    $this->pageCacheKillSwitch->trigger();
    $event->setResponse($response);
  }

  /**
   * Mitigate Autologout helper function.
   *
   * Check to see if the autologout module is going to log us out on this
   * request or within 10 seconds. If so, then reauthenticate via the SSO
   * identity provider.
   */
  protected function mitigateAutoLogout(RequestEvent $event) {
    $store = $this->tempStoreFactory->get('sso_trigger');

    // If we're already trying to re-authenticate, then do that now.
    if ($store->get('sso_trigger.reauth')) {
      // This is a bit of a kludge, but unsetting the autologout_last session
      // variable seems to be the only way to prevent an infinite redirect loop.
      unset($_SESSION['autologout_last']);
      $store->delete('sso_trigger.reauth');

      // Destroy the session and trigger SSO.
      $this->getLogger('sso_trigger')->notice('Destroying Drupal session');
      $this->sessionManager->destroy();

      $this->account->setAccount(new AnonymousUserSession());
      $this->getLogger('sso_trigger')->notice('Triggering SSO');
      $this->triggerSso($event);
      return;
    }

    // If the autologout module is going to log us out on this request
    // or within 10 seconds, then reauthenticate via the SSO identify
    // provider.
    $ttl = $this->autologoutManager->getRemainingTime();
    // If autologout is going to log us out within 10 seconds.
    if ($ttl <= 10) {
      if (!$store->get('sso_trigger.reauth')) {
        $store->set('sso_trigger.reauth', TRUE);
        $this->getLogger('sso_trigger')->notice('TTL is @t seconds. Forcing re-auth.', ['@t' => $ttl]);
        $this->authManager->logout($this->pathStackService->getPath());
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[KernelEvents::REQUEST][] = ['checkAuthStatus', 110];
    return $events;
  }

}
